@@ -4,7 +4,7 @@ import time
 import sys
 import socket
 import errno
-import urllib
+import weakref
 import traceback
 
 try:
@@ -12,11 +12,10 @@ try:
 except ImportError:  # Windows
     import getpass as pwd
 
-try:
-    from hashlib import sha1
-except ImportError:  # Python 2.4 or earlier
-    from sha import new as sha1
-
+from supervisor.compat import urllib
+from supervisor.compat import sha1
+from supervisor.compat import as_bytes
+from supervisor.compat import as_string
 from supervisor.medusa import asyncore_25 as asyncore
 from supervisor.medusa import http_date
 from supervisor.medusa import http_server
@@ -50,18 +49,19 @@ class deferring_chunked_producer:
             if data is NOT_DONE_YET:
                 return NOT_DONE_YET
             elif data:
-                return '%x\r\n%s\r\n' % (len(data), data)
+                s = '%x' % len(data)
+                return as_bytes(s) + b'\r\n' + data + b'\r\n'
             else:
                 self.producer = None
                 if self.footers:
-                    return '\r\n'.join(['0'] + self.footers) + '\r\n\r\n'
+                    return b'\r\n'.join([b'0'] + self.footers) + b'\r\n\r\n'
                 else:
-                    return '0\r\n\r\n'
+                    return b'0\r\n\r\n'
         else:
-            return ''
+            return b''
 
 class deferring_composite_producer:
-    "combine a fifo of producers into one"
+    """combine a fifo of producers into one"""
     def __init__ (self, producers):
         self.producers = producers
         self.delay = 0.1
@@ -77,7 +77,7 @@ class deferring_composite_producer:
             else:
                 self.producers.pop(0)
         else:
-            return ''
+            return b''
 
 
 class deferring_globbing_producer:
@@ -89,7 +89,7 @@ class deferring_globbing_producer:
 
     def __init__ (self, producer, buffer_size=1<<16):
         self.producer = producer
-        self.buffer = ''
+        self.buffer = b''
         self.buffer_size = buffer_size
         self.delay = 0.1
 
@@ -99,11 +99,14 @@ class deferring_globbing_producer:
             if data is NOT_DONE_YET:
                 return NOT_DONE_YET
             if data:
-                self.buffer = self.buffer + data
+                try:
+                    self.buffer = self.buffer + data
+                except TypeError:
+                    self.buffer = as_bytes(self.buffer) + as_bytes(data)
             else:
                 break
         r = self.buffer
-        self.buffer = ''
+        self.buffer = b''
         return r
 
 
@@ -129,10 +132,10 @@ class deferring_hooked_producer:
                 self.producer = None
                 self.function (self.bytes)
             else:
-                self.bytes = self.bytes + len(result)
+                self.bytes += len(result)
             return result
         else:
-            return ''
+            return b''
 
 
 class deferring_http_request(http_server.http_request):
@@ -289,11 +292,11 @@ class deferring_http_request(http_server.http_request):
             key,value=header.split(":",1)
             key=key.lower()
             value=value.strip()
-            if header2env.has_key(key) and value:
+            if key in header2env and value:
                 env[header2env.get(key)]=value
             else:
                 key='HTTP_%s' % ("_".join(key.split( "-"))).upper()
-                if value and not env.has_key(key):
+                if value and key not in env:
                     env[key]=value
         return env
 
@@ -310,14 +313,14 @@ class deferring_http_request(http_server.http_request):
         else:
             protocol = 'http'
 
-        if environ.has_key('HTTP_HOST'):
+        if 'HTTP_HOST' in environ:
             host = environ['HTTP_HOST'].strip()
             hostname, port = urllib.splitport(host)
         else:
             hostname = environ['SERVER_NAME'].strip()
             port = environ['SERVER_PORT']
 
-        if (port is None or default_port[protocol] == port):
+        if port is None or default_port[protocol] == port:
             host = hostname
         else:
             host = hostname + ':' + port
@@ -362,9 +365,9 @@ class deferring_http_channel(http_server.http_channel):
                         self.producer_fifo.pop()
                         self.close()
                     return
-                elif isinstance(p, str):
+                elif isinstance(p, bytes):
                     self.producer_fifo.pop()
-                    self.ac_out_buffer = self.ac_out_buffer + p
+                    self.ac_out_buffer += p
                     return
 
                 data = p.more()
@@ -389,8 +392,11 @@ class deferring_http_channel(http_server.http_channel):
         if self.current_request:
             self.current_request.found_terminator()
         else:
-            header = self.in_buffer
-            self.in_buffer = ''
+            # we convert the header to text to facilitate processing.
+            # some of the underlying APIs (such as splitquery)
+            # expect text rather than bytes.
+            header = as_string(self.in_buffer)
+            self.in_buffer = b''
             lines = header.split('\r\n')
 
             # --------------------------------------------------
@@ -418,12 +424,12 @@ class deferring_http_channel(http_server.http_channel):
             rpath, rquery = http_server.splitquery(uri)
             if '%' in rpath:
                 if rquery:
-                    uri = http_server.unquote (rpath) + '?' + rquery
+                    uri = http_server.unquote(rpath) + '?' + rquery
                 else:
-                    uri = http_server.unquote (rpath)
+                    uri = http_server.unquote(rpath)
 
-            r = deferring_http_request (self, request, command, uri, version,
-                                         header)
+            r = deferring_http_request(self, request, command, uri, version,
+                                       header)
             self.request_counter.increment()
             self.server.total_requests.increment()
 
@@ -522,7 +528,6 @@ class supervisor_af_inet_http_server(supervisor_http_server):
         self.prebind(sock, logger_object)
         self.bind((ip, port))
 
-        host, port = self.socket.getsockname()
         if not ip:
             self.log_info('Computing default hostname', 'warning')
             hostname = socket.gethostname()
@@ -591,7 +596,7 @@ class supervisor_af_unix_http_server(supervisor_http_server):
                 else:
                     try:
                         os.chown(socketname, sockchown[0], sockchown[1])
-                    except OSError, why:
+                    except OSError as why:
                         if why.args[0] == errno.EPERM:
                             msg = ('Not permitted to chown %s to uid/gid %s; '
                                    'adjust "sockchown" value in config file or '
@@ -621,7 +626,7 @@ class supervisor_af_unix_http_server(supervisor_http_server):
         s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
         try:
             s.connect(socketname)
-            s.send("GET / HTTP/1.0\r\n\r\n")
+            s.send(as_bytes("GET / HTTP/1.0\r\n\r\n"))
             s.recv(1)
             s.close()
         except socket.error:
@@ -631,7 +636,7 @@ class supervisor_af_unix_http_server(supervisor_http_server):
 
 class tail_f_producer:
     def __init__(self, request, filename, head):
-        self.request = request
+        self.request = weakref.ref(request)
         self.filename = filename
         self.delay = 0.1
 
@@ -649,7 +654,7 @@ class tail_f_producer:
             newsz = self._fsize()
         except (OSError, ValueError):
             # file descriptor was closed
-            return ''
+            return b''
         bytes_added = newsz - self.sz
         if bytes_added < 0:
             self.sz = 0
@@ -732,10 +737,11 @@ class logtail_handler:
         logfile = getattr(process.config, '%s_logfile' % channel, None)
 
         if logfile is None or not os.path.exists(logfile):
-            # XXX problematic: processes that don't start won't have a log
-            # file and we probably don't want to go into fatal state if we try
-            # to read the log of a process that did not start.
-            request.error(410) # gone
+            # we return 404 because no logfile is a temporary condition.
+            # if the process has never been started, no logfile will exist
+            # on disk.  a logfile of None is also a temporay condition,
+            # since the config file can be reloaded.
+            request.error(404) # not found
             return
 
         mtime = os.stat(logfile)[stat.ST_MTIME]
@@ -743,6 +749,9 @@ class logtail_handler:
         request['Content-Type'] = 'text/plain;charset=utf-8'
         # the lack of a Content-Length header makes the outputter
         # send a 'Transfer-Encoding: chunked' response
+        request['X-Accel-Buffering'] = 'no'
+        # tell reverse proxy server (e.g., nginx) to disable proxy buffering
+        # (see also http://nginx.org/en/docs/http/ngx_http_proxy_module.html#proxy_buffering)
 
         request.push(tail_f_producer(request, logfile, 1024))
 
@@ -766,7 +775,10 @@ class mainlogtail_handler:
         logfile = self.supervisord.options.logfile
 
         if logfile is None or not os.path.exists(logfile):
-            request.error(410) # gone
+            # we return 404 because no logfile is a temporary condition.
+            # even if a log file of None is configured, the config file
+            # may be reloaded, and the new config may have a logfile.
+            request.error(404) # not found
             return
 
         mtime = os.stat(logfile)[stat.ST_MTIME]
@@ -799,9 +811,9 @@ def make_http_servers(options, supervisord):
         else:
             raise ValueError('Cannot determine socket type %r' % family)
 
-        from xmlrpc import supervisor_xmlrpc_handler
-        from xmlrpc import SystemNamespaceRPCInterface
-        from web import supervisor_ui_handler
+        from supervisor.xmlrpc import supervisor_xmlrpc_handler
+        from supervisor.xmlrpc import SystemNamespaceRPCInterface
+        from supervisor.web import supervisor_ui_handler
 
         subinterfaces = []
         for name, factory, d in options.rpcinterface_factories:
@@ -875,10 +887,10 @@ class encrypted_dictionary_authorizer:
 
     def authorize(self, auth_info):
         username, password = auth_info
-        if self.dict.has_key(username):
+        if username in self.dict:
             stored_password = self.dict[username]
             if stored_password.startswith('{SHA}'):
-                password_hash = sha1(password).hexdigest()
+                password_hash = sha1(as_bytes(password)).hexdigest()
                 return stored_password[5:] == password_hash
             else:
                 return stored_password == password
