@@ -46,6 +46,7 @@ from supervisor.datatypes import InetStreamSocketConfig
 from supervisor.datatypes import UnixStreamSocketConfig
 from supervisor.datatypes import url
 from supervisor.datatypes import Automatic
+from supervisor.datatypes import Syslog
 from supervisor.datatypes import auto_restart
 from supervisor.datatypes import profile_options
 
@@ -101,6 +102,7 @@ class Options:
         self.attr_priorities = {}
         self.require_configfile = require_configfile
         self.add(None, None, "h", "help", self.help)
+        self.add(None, None, "?", None, self.help)
         self.add("configfile", None, "c:", "configuration=")
 
         here = os.path.dirname(os.path.dirname(sys.argv[0]))
@@ -404,7 +406,6 @@ class ServerOptions(Options):
     passwdfile = None
     nodaemon = None
     silent = None
-    environment = None
     httpservers = ()
     unlink_pidfile = False
     unlink_socketfiles = False
@@ -501,7 +502,11 @@ class ServerOptions(Options):
         else:
             logfile = section.logfile
 
-        self.logfile = normalize_path(logfile)
+        if logfile != 'syslog':
+            # if the value for logfile is "syslog", we don't want to
+            # normalize the path to something like $CWD/syslog.log, but
+            # instead use the syslog service.
+            self.logfile = normalize_path(logfile)
 
         if self.pidfile:
             pidfile = self.pidfile
@@ -538,8 +543,6 @@ class ServerOptions(Options):
         # self.serverurl may still be None if no servers at all are
         # configured in the config file
 
-        self.identifier = section.identifier
-
     def process_config(self, do_usage=True):
         Options.process_config(self, do_usage=do_usage)
 
@@ -651,6 +654,11 @@ class ServerOptions(Options):
         environ_str = get('environment', '')
         environ_str = expand(environ_str, expansions, 'environment')
         section.environment = dict_of_key_value_pairs(environ_str)
+
+        # extend expansions for global from [supervisord] environment definition
+        for k, v in section.environment.items():
+            self.environ_expansions['ENV_%s' % k ] = v
+
         # Process rpcinterface plugins before groups to allow custom events to
         # be registered.
         section.rpcinterface_factories = self.get_plugins(
@@ -958,17 +966,21 @@ class ServerOptions(Options):
             environment = dict_of_key_value_pairs(
                 expand(environment_str, expansions, 'environment'))
 
+            # extend expansions for process from [program:x] environment definition
+            for k, v in environment.items():
+                expansions['ENV_%s' % k] = v
+
             directory = get(section, 'directory', None)
 
             logfiles = {}
 
             for k in ('stdout', 'stderr'):
-                n = '%s_logfile' % k
-                lf_val = get(section, n, Automatic)
+                lf_key = '%s_logfile' % k
+                lf_val = get(section, lf_key, Automatic)
                 if isinstance(lf_val, basestring):
-                    lf_val = expand(lf_val, expansions, n)
+                    lf_val = expand(lf_val, expansions, lf_key)
                 lf_val = logfile_name(lf_val)
-                logfiles[n] = lf_val
+                logfiles[lf_key] = lf_val
 
                 bu_key = '%s_logfile_backups' % k
                 backups = integer(get(section, bu_key, 10))
@@ -982,11 +994,21 @@ class ServerOptions(Options):
                 syslog = boolean(get(section, sy_key, False))
                 logfiles[sy_key] = syslog
 
+                # rewrite deprecated "syslog" magic logfile into the equivalent
+                # TODO remove this in a future version
+                if lf_val is Syslog:
+                    self.parse_warnings.append(
+                        'For [%s], %s=syslog but this is deprecated and will '
+                        'be removed.  Use %s=true to enable syslog instead.' % (
+                        section, lf_key, sy_key))
+                    logfiles[lf_key] = lf_val = None
+                    logfiles[sy_key] = True
+
                 if lf_val is Automatic and not maxbytes:
                     self.parse_warnings.append(
                         'For [%s], AUTO logging used for %s without '
                         'rollover, set maxbytes > 0 to avoid filling up '
-                        'filesystem unintentionally' % (section, n))
+                        'filesystem unintentionally' % (section, lf_key))
 
             if redirect_stderr:
                 if logfiles['stderr_logfile'] not in (Automatic, None):
@@ -1568,9 +1590,6 @@ class ServerOptions(Options):
             data = b''
         return data
 
-    def process_environment(self):
-        os.environ.update(self.environment or {})
-
     def chdir(self, dir):
         os.chdir(dir)
 
